# -*- coding: utf-8 -*-
"""Global settings for rpi2caster. Acts as a single point of truth (SPOT)
for every application-wide setting."""
import io
import configparser
from click import get_app_dir
from . import text_ui as UI
from .constants import TRUE_ALIASES, FALSE_ALIASES, SIGNALS


def get_config(section_name, option_name, default_value, datatype=str):
    """get_config:

    Gets a value for a given parameter from a given name.
    Returns:
      -int - if the value is numeric,
      -boolean - if the value was in one of the true or false aliases,
      -string otherwise,
      -None if there is no option or section of that name.
    """
    # Search user and global config paths
    for path in CONFIG_PATHS:
        try:
            with io.open(path, 'r'):
                cfg = configparser.ConfigParser()
                cfg.read(path)
                option_value = str(cfg.get(section_name, option_name))
                # Go further to processing
                break
        except (IOError, FileNotFoundError,
                configparser.NoSectionError, configparser.NoOptionError):
            # No option found in this config file - tr
            continue
    else:
        # Ran through all config files and found no option
        # In this case, use default...
        return default_value
    # We now have the value - let's convert it to desired datatype
    if datatype == list:
        retval = option_value.split(',')
    elif datatype == int and str(option_value).lower().startswith('0x'):
        # Value is a hexstring: 0x or 0X
        try:
            retval = int(option_value, 16)
        except (ValueError, TypeError):
            # Conversion failed
            retval = default_value
    elif option_value.lower() in ('none', 'null'):
        retval = None
    elif datatype == bool and str(option_value).lower() in TRUE_ALIASES:
        retval = True
    elif datatype == bool and str(option_value).lower() in FALSE_ALIASES:
        retval = False
    else:
        retval = datatype(option_value)
    return retval


# Conffile path
APPDIR = get_app_dir('rpi2caster', force_posix=True, roaming=True)
CONFIG_PATHS = [APPDIR + '/rpi2caster.conf',
                '/etc/rpi2caster/rpi2caster.conf', '/etc/rpi2caster.conf']
# SQLite3 database path
DATABASE_PATHS = [APPDIR + '/rpi2caster.db',
                  '/var/local/rpi2caster/rpi2caster.db',
                  '/var/rpi2caster/rpi2caster.db']
# Line length / galley width default value
DEFAULT_MEASURE = get_config('Preferences', 'default_measure', 25, int)
DEFAULT_UNIT = get_config('Preferences', 'measurement_unit', 'cc')
# Interface settings
# Inputs for the casting interface
SIGNALS_ARRANGEMENT = get_config('Interface', 'signals', SIGNALS, list)
EMERGENCY_STOP_GPIO = get_config('Interface', 'stop_gpio', 22, int)
SENSOR_GPIO = get_config('Interface', 'sensor_gpio', 17, int)
# Default hardware interface output data
MCP0 = get_config('Interface', 'MCP0', 0x20, int)
MCP1 = get_config('Interface', 'MCP1', 0x21, int)
PIN_BASE = get_config('Interface', 'pin_base', 65, int)
SENSOR = get_config('Interface', 'sensor', 'simulation')
OUTPUT = get_config('Interface', 'output', 'simulation')
# Now choose the simulation or real sensor/interface...
SELECT = get_config('Preferences', 'choose_backend', False, bool)
if SELECT and SENSOR != 'simulation' and OUTPUT != 'simulation':
    if not UI.confirm('Use caster? (no = simulation mode)', True):
        SENSOR = 'simulation'
        OUTPUT = 'simulation'



# -*- coding: utf-8 -*-
"""Parallel port input and output driver"""
import time
from parallel import Parallel
from .monotype import SimulationSensor, SimulationOutput
from .exceptions import return_to_menu
from .global_config import UI
from .constants import SIGNALS
from .helpers import singleton

NEW_SIGNALS = ['O15'] + SIGNALS[:-1]
MAP = dict(zip(NEW_SIGNALS, (2 ** x for x in range(31, -1, -1))))


@singleton
class ParallelOutputDriver(SimulationOutput):
    """Output driver for parallel port. Sends four bytes in sequence:
    byte0: O N M L K J I H
    byte1: G F S E D 0075 C B
    byte2: A 1 2 3 4 5 6 7
    byte3: 8 9 10 11 12 13 14 0005
    """
    def __init__(self):
        super().__init__()
        self.working = False
        self.port = None
        self.lock = False

    def __enter__(self):
        if not self.lock:
            try:
                self.lock = True
                return self
            except OSError:
                UI.pause('Cannot set up the parallel port. Exiting...')

    def __exit__(self, *_, **__):
        self.lock = False
        # Release the parallel port
        self.port = None

    def valves_on(self, signals_list=None):
        """Activate the valves"""
        if signals_list:
            number = sum(MAP.get(signal, 0) for signal in signals_list)
            # Split it to four bytes sent in sequence
            byte0 = (number >> 24) & 0xff
            byte1 = (number >> 16) & 0xff
            byte2 = (number >> 8) & 0xff
            byte3 = number & 0xff
        else:
            # Turn off the valves
            byte0 = byte1 = byte2 = byte3 = 0x00
        # Display debug information about the bytes
        UI.debug_info('%s %s %s %s' % (format(byte0, 'b').zfill(8),
                                       format(byte1, 'b').zfill(8),
                                       format(byte2, 'b').zfill(8),
                                       format(byte3, 'b').zfill(8)))
        for byte in (byte0, byte1, byte2, byte3):
            self._send(byte)

    def valves_off(self):
        """Deactivate the valves - actually, do nothing"""
        pass

    def initialize(self):
        """Startup sequence for the parallel port"""
        if not self.port:
            try:
                self.port = Parallel()
            except (FileNotFoundError, IOError, OSError):
                UI.pause('ERROR: Cannot access the parallel port! '
                         'Check your hardware and OS configuration...')
                return_to_menu()
        if not self.working:
            self._init_on()
            self._init_off()
            UI.display('\nPress the button on the interface...')
            self._wait_until_not_busy()
            self.working = True

    def _send(self, data):
        """Send the codes through the data port"""
        # Wait until we can send the codes
        self._wait_until_not_busy()
        self._set_data(data)
        # Signal that we finished and wait until interface acknowledges
        self._strobe_on()
        self._wait_until_busy()
        self._strobe_off()
        self._wait_until_not_busy()

    def _set_data(self, data):
        "Set the lines on the data port"""
        self.port.setData(data)

    def _wait_until_busy(self):
        """Wait until busy goes ON"""
        while self.port.getInBusy():
            pass

    def _wait_until_not_busy(self):
        """Wait until busy goes OFF"""
        while not self.port.getInBusy():
            pass

    def _strobe_on(self):
        """Send the strobe signal"""
        # Negative logic!
        self.port.setDataStrobe(False)
        # 5us sleep
        time.sleep(0.000005)

    def _strobe_off(self):
        """Send the strobe signal"""
        # Negative logic!
        self.port.setDataStrobe(True)

    def _init_on(self):
        """Sends the init signal"""
        # Negative logic!
        self.port.setInitOut(False)

    def _init_off(self):
        """Sends the init signal"""
        # Negative logic!
        self.port.setInitOut(True)


@singleton
class ParallelSensor(SimulationSensor):
    """Parallel port sensor driver"""
    def __init__(self):
        super().__init__()

    def check_if_machine_is_working(self):
        """Reset the interface if needed and go on"""
        ParallelOutputDriver().initialize()
        UI.pause('Turn on the machine... (or ctrl-C to abort)')

    def wait_for(self, *args, **kw):
        """Do nothing"""
        pass



# -*- coding: utf-8 -*-
"""Command-line interface functions for rpi2caster"""

# IMPORTS for text user interface
import io
import os
import readline
import glob

from . import exceptions as e
from . import constants as c

# Whether the debug mode is on (can be changed by setting module's attribute)
DEBUG_MODE = False
# Some standard prompts
MSG_MENU = '[Enter] to go back to main menu...'
MSG_CONTINUE = '[Enter] to continue...'


def menu(options, header='', footer='', no_debug=False):
    """menu(options=[(name1, long1, func1), (name2, long2, func2)...],
                    header=foo,
                    footer=bar,
                    no_debug=False):

    A menu which takes four arguments:
    options list - contains tuples for each option:
        (description, long_description, function)
    header - string to be displayed above,
    footer - string to be displayed below,
    no_debug - True if we don't want to tell the user
        that the program is in debugging mode.

    After choice is made, return the command.

    Set up vars for conditional statements,
    and lists for appending new items.

    choices - options to be entered by user
    """
    # Clear the screen, display header and add two empty lines
    clear()
    if header:
        print(header, end='\n\n')
    # Get the first option - this will be listed last
    try:
        (zero_function, zero_desc, zero_long_desc) = options[0]
        functions = [zero_function]
    except IndexError:
        raise e.MenuLevelUp
    # Display all the options
    # Tab indent, option number, option name (not processing option 0 yet!)
    for i, (function, desc, long_desc) in enumerate(options):
        if i > 0:
            functions.append(function)
            if long_desc:
                long_desc = '\n\t\t %s ' % long_desc
            print('\t %i : %s%s\n' % (i, desc, long_desc))
    # Option 0 is displayed last, add some whitespace around it
    if zero_desc:
        print('\n\t %i : %s \n\t\t %s \n' % (0, zero_desc, zero_long_desc))
    # Print footer, if defined
    if footer:
        print(footer, end='\n\n')
    if DEBUG_MODE and not no_debug:
        print('The program is now in debugging mode!', end='\n\n')
    # Add an empty line to separate prompt
    print('\n')
    # Ask for user input
    choice_number = -1
    # Get only available options and exclude non-numeric strings
    while choice_number not in range(len(options)):
        # Wait until user enters proper data
        your_choice = input('Your choice: ')
        try:
            choice_number = int(your_choice)
        except ValueError:
            # Entered anything non-digit - repeat
            your_choice = ''
    # At last, we have chosen a valid option...
    # Return a corresponding value - which is option
    return functions[choice_number]


def clear():
    """Clears the screen"""
    os.system('clear')


def display(*args, **kwargs):
    """Displays info for the user - print all in one line"""
    print(*args, **kwargs)


def display_header(text, symbol='-'):
    """Displays a header banner"""
    dash_line = symbol * len(text)
    print('\n' + dash_line + '\n' + text + '\n' + dash_line + '\n')


def display_parameters(data):
    """Displays the parameters by section (given as a dictionary):
    {header1: [(param1_val1, param1_desc1), (param1_val2, param1_desc2)...],
     header2: [(param2_val1, param2_desc1), (param2_val2, param2_desc2)...]...}
     a section will be displayed if there are parameters to display;
     a parameter will be displayed if its value evaluates to True."""
    # {header: [(par1, desc1), (par2, desc2)], header2: [(...)]}
    for key in data:
        parameters = '\n'.join(['%s: %s' % (desc, value)
                                for (value, desc) in data[key] if value])
        if parameters:
            display_header(key)
            print(parameters)


def debug_info(*args, **kwargs):
    """Prints debug messages to screen if in debug mode"""
    if DEBUG_MODE:
        print('DEBUG: ', *args, **kwargs)


def debug_enter_data(prompt, datatype=str):
    """For debug-specific data"""
    if DEBUG_MODE:
        return enter_data('DEBUG: ' + prompt + ' : ', datatype)


def debug_pause(msg1='', msg2=MSG_CONTINUE):
    """For debug confirmations"""
    if DEBUG_MODE:
        input('DEBUG: ' + msg1 + ' - ' + msg2)
        print('\n')


def pause(msg1='', msg2=MSG_CONTINUE):
    """Waits until user presses return"""
    input(msg1 + '\n' + msg2)
    print('\n')


def enter_data(prompt, datatype=str):
    """Enter a value and convert it to the specific datatype"""
    value = ''
    while not value:
        value = input(prompt + ' : ')
        print()
        try:
            value = datatype(value)
        except ValueError:
            print('Incorrect value or data type!')
            value = ''
    return value


def enter_data_or_blank(prompt, datatype=str):
    """Enter a value or leave blank, try to convert to the specified datatype
    """
    while True:
        value = input(prompt + ' : ')
        print()
        if not value and datatype in (int, float, bool):
            return datatype(0)
        elif not value:
            return value
        try:
            return datatype(value)
        except ValueError:
            print('Incorrect value or data type!')


def enter_data_or_default(prompt, default=1, datatype=str):
    """Enter a value and return default if not given anything,
    or try to convert to the specified datatype."""
    while True:
        value = input(prompt + ' (default: %s) : ' % default)
        print()
        if not value:
            return default
        try:
            return datatype(value)
        except ValueError:
            print('Incorrect value or data type!')


def enter_data_or_exception(prompt, exception=ValueError, datatype=str):
    """Enter a value and raise an exception if it is blank,
    or try to convert to the specified datatype."""
    while True:
        value = input(prompt + ' : ')
        print()
        if not value:
            raise exception
        try:
            return datatype(value)
        except ValueError:
            print('Incorrect value or data type!')


def tab_complete(text, state):
    """Enables tab key auto-completion when you enter the filename."""
    return (glob.glob(text+'*')+[None])[state]


def enter_input_filename():
    """Allows to enter the input filename and checks if it is readable.
    Repeats until proper filename or nothing is given.
    Raises ReturnToMenu if filename not specified."""
    # Set readline parameters
    readline.set_completer_delims(' \t\n;')
    readline.parse_and_bind('tab: complete')
    readline.set_completer(tab_complete)
    # Enter the input filename; check if the file is readable
    while True:
        prompt = 'Enter the input file name (leave blank to abort)'
        filename = enter_data_or_exception(prompt, e.ReturnToMenu)
        filename = os.path.realpath(filename)
        try:
            with io.open(filename, 'r'):
                return filename
        except (IOError, FileNotFoundError):
            print('Wrong filename or file not readable!\n')


def enter_output_filename():
    """Allows user to enter output filename (without checking if readable).
    Raises ReturnToMenu if filename not specified."""
    # Set readline parameters
    readline.set_completer_delims(' \t\n;')
    readline.parse_and_bind('tab: complete')
    readline.set_completer(tab_complete)
    # Enter the output filename; no check here
    prompt = 'Enter the input file name (leave blank to abort)'
    filename = enter_data_or_exception(prompt, e.ReturnToMenu)
    return os.path.realpath(filename)


def simple_menu(message, options):
    """A simple menu where user is asked what to do.
    Wrong choice points back to the menu.

    Message: prompt - string displayed on screen;
    options: a dict {ans1:opt1, ans2:opt2...}.
    """
    ans = ''
    while True:
        ans = input(message)
        print('\n')
        if ans in options:
            return options[ans]
        elif ans.lower() in options:
            return options[ans.lower()]
        elif ans.upper() in options:
            return options[ans.upper()]
        else:
            pass


def confirm(question, default=None):
    """Asks a simple question with yes or no answers.
    Returns True for yes and False for no."""
    options = {'Y': True, 'N': False}
    def_string = ''
    if default is True or default is False:
        options[''] = default
        def_string = ', default: %s' % (default and 'Y' or 'N')
    return simple_menu('%s [Y / N%s]: ' % (question, def_string), options)


def display_diecase_layout(diecase):
    """Shows a layout for a given diecase ID, unit values for its
    assigned wedge, or the typical S5 if not specified."""
    def get_char(matrix):
        """Modifies matrix char for displaying"""
        # Style modifiers for displaying roman, bold, italic,
        # smallcaps, inferior, superior
        style_modifiers = {'r': '', 'b': '*', 'i': '/',
                           's': '·', 'u': '_', 'l': '^'}
        spaces_symbols = {'_': '▣', ' ': '□', '': ' '}
        fmt = ''.join([style_modifiers.get(s, '') for s in matrix.styles])
        if len(fmt) > 2:
            fmt = '#'
        return spaces_symbols.get(matrix.char, fmt + matrix.char)

    def get_width(column):
        """Get the column width"""
        return column_widths.get(column, 4)

    def get_unit_value(mat):
        """Get the unit value string for a character, if the width
        overrides the wedge default"""
        show_units = (mat.units != wedge_row_units and
                      mat.char and not mat.isspace())
        if show_units:
            return str(mat.units)
        else:
            return ''

    col_numbers = c.COLUMNS_15
    row_numbers = [x for x in range(1, 16)]
    # Set the initial column width value to 4
    column_widths = {}
    # Now loop over the matrices and adjust the displayed table
    for mat in diecase:
        if '16' in mat.code:
            col_numbers = c.COLUMNS_17
            row_numbers = [x for x in range(1, 17)]
            break
        elif mat.column in ('NI', 'NL'):
            col_numbers = c.COLUMNS_17
        # Update column width if displayed character turns out too long
        if len(get_char(mat)) + 2 > column_widths.get(mat.column, 4):
            column_widths[mat.column] = len(get_char(mat)) + 2
    # Generate a header with column numbers
    # Widths are variable and depend on text field width
    fields = [col.center(get_width(col)) for col in col_numbers]
    header = ' Row %s Units ' % ''.join(fields)
    # "-----" line in the table
    separator = '—' * len(header)
    # Initialize the displayed layout
    table = [header, separator]
    # Process each row
    for row in row_numbers:
        items = []
        units_row = []
        wedge_row_units = diecase.wedge[row]
        for col in col_numbers:
            width = get_width(col)
            for mat in diecase:
                if mat.code == '%s%s' % (col, row):
                    items.append(get_char(mat).center(width))
                    units_row.append(get_unit_value(mat).center(width))
                    # No need to iterate further
                    break
        items = ['|', str(row).center(3), '|', ''.join(items),
                 '|%s|' % str(wedge_row_units).center(5)]
        units_row = ['|', ' ' * 3, '|', ''.join(units_row), '|', ' ' * 5, '|']
        table.append(''.join(items))
        table.append(''.join(units_row))
        # table.append(empty_row)
    # Add the header at the bottom
    table.extend([separator, header])
    # We can display it now
    print('\nDiecase ID: %s  -   assigned stopbar/wedge: %s\n'
          % (diecase, diecase.wedge))
    for row in table:
        print(row)
    # Explanation of symbols
    print('\nExplanation:', '□ = low space, ▣ = high space',
          '*a = bold, /a = italic, ·a = small caps',
          '_a = subscript (inferior), ^a = superscript (superior)',
          '# - matrix assigned to more than two styles', sep='\n')


def edit_diecase_layout(diecase):
    """Edits a matrix case layout, row by row, matrix by matrix.
    Allows to enter a position to be edited. """
    def swap(command):
        """Swap two matrices based on command"""
        # Process the command string (uppercase)
        command = command.replace('SWAP', '').strip()
        code1, code2 = command.split(',', 1)
        code1, code2 = code1.strip(), code2.strip()
        # Look for matrices
        matches1 = [mat for mat in diecase.matrices if mat.code == code1]
        matches2 = [mat for mat in diecase.matrices if mat.code == code2]
        if matches1 and matches2:
            mat1, mat2 = matches1[0], matches2[0]
            mat1.code, mat2.code = mat2.code, mat1.code

    def edit(mat):
        """Edit a matrix"""
        clear()
        display_diecase_layout(diecase)
        mat.edit()

    def all_rows_mode():
        """Row-by-row editing - all cells in row 1, then 2 etc."""
        for mat in diecase:
            edit(mat)

    def all_columns_mode():
        """Column-by-column editing - all cells in column NI, NL, A...O"""
        # Rearrange the layout so that it's read column by column
        iter_mats = (mat for col in c.COLUMNS_17
                     for mat in diecase if mat.column == col)
        for mat in iter_mats:
            edit(mat)

    def single_row_mode(row):
        """Edits matrices found in a single row"""
        iter_mats = (mat for mat in diecase if mat.row == row)
        for mat in iter_mats:
            edit(mat)

    def single_column_mode(column):
        """Edits matrices found in a single column"""
        iter_mats = (mat for mat in diecase if mat.column == column)
        for mat in iter_mats:
            edit(mat)

    # Map unit values to rows
    # If the layout is empty, we need to initialize it
    prompt = ('Enter row number to edit all mats in a row,\n'
              'column number to edit all mats in a column,\n'
              'matrix coordinates to edit a single matrix,\n'
              'or choose edit mode: AR - all matrices row by row, '
              'AC - all matrices column by column.'
              '\nYou can swap two mats by entering: "swap coords1, coords2".'
              '\nYour choice (or leave blank to exit) : ')
    while True:
        print('\nCurrent diecase layout:\n')
        display_diecase_layout(diecase)
        try:
            ans = input(prompt).upper()
            if ans == 'AR':
                all_rows_mode()
            elif ans == 'AC':
                all_columns_mode()
            elif ans in c.COLUMNS_17:
                single_column_mode(ans)
            elif ans in [str(x) for x in range(1, 17)]:
                single_row_mode(int(ans))
            elif ans.startswith('SWAP'):
                swap(ans)
            elif ans:
                try:
                    mats = [mat for mat in diecase if mat.code == ans.upper()]
                    edit(mats[0])
                except (IndexError, TypeError, AttributeError):
                    # Loop over again
                    pass
            else:
                return diecase.matrices
        except (IndexError, KeyboardInterrupt, TypeError, AttributeError):
            pass



# -*- coding: utf-8 -*-
"""Line length and character width module"""
from .global_config import UI, DEFAULT_MEASURE, DEFAULT_UNIT
from .wedge_data import Wedge


class Measure(object):
    """Chooses and represents a line length"""
    default_unit = DEFAULT_UNIT
    default_value = DEFAULT_MEASURE
    symbols = ['pc', 'pt', 'Pp', 'pp', 'cc', 'dd', 'cf', 'ff',
               'cm', 'mm', 'in', '"']
    units = {'pc': 12.0, 'pt': 1.0,
             'Pp': 12*0.166/0.1667, 'pp': 0.166/0.1667,
             'cc': 12*0.1776/0.1667, 'dd': 0.1776/0.1667,
             'cf': 12*0.1628/0.1667, 'ff': 0.1628/0.1667,
             'cm': 0.3937*72, 'mm': 0.03937*72, '"': 72.0, 'in': 72.0}

    def __init__(self, value=default_value, unit=default_unit, manual=True):
        try:
            if manual or not value:
                raise ValueError
            string = '%s%s' % (value, unit)
            self.points = parse_value(string)
        except (TypeError, ValueError, AttributeError):
            self.points = enter(value, unit)

    def __str__(self):
        return '%s%s' % (self.type_units, Measure.default_unit)

    @property
    def type_units(self):
        """Get a value in default units specified in configuration"""
        # Get the coefficient for calculation
        factor = 1 / Measure.units.get(Measure.default_unit, 1)
        return round(self.points * factor, 2)

    def ems(self, type_size=12):
        """Gets the number of ems of specified type size"""
        return round(self.points / type_size, 2)

    def monotype_units(self, wedge=Wedge()):
        """Calculates the line length in wedge's units"""
        return round(self.points / wedge.units_to_points_factor, 2)


def enter(value=0, unit=Measure.default_unit):
    """Enter the line length, choose measurement units
    (for e.g. British or European measurement system).
    Return length in DTP points."""
    def unit_help():
        """Display a help text about available units"""
        text = ('\n\nAvailable units:\n'
                '    dd - European Didot point,\n'
                '    cc - cicero (=12dd, .1776"),\n'
                '    ff - Fournier point,\n'
                '    cf - Fournier cicero (=12ff, .1628"),\n'
                '    pp - TeX / US printer\'s pica point,\n'
                '    Pp - TeX / US printer\'s pica (=12pp, .1660"),\n'
                '    pt - DTP / PostScript point = 1/72",\n'
                '    pc - DTP / PostScript pica (=12pt, .1667"),\n'
                '    ", in - inch;   '
                'mm - millimeter;   cm - centimeter\n\n')
        UI.display(text)

    prompt = 'Enter the length/width value and unit (or "?" for help)'
    value = value or Measure.default_value
    default_value = '%s%s' % (value, unit)
    while True:
        # If 0, use default
        raw_string = UI.enter_data_or_default(prompt, default_value)
        if '?' in raw_string:
            # Display help message and start again
            unit_help()
            continue
        try:
            return parse_value(raw_string)
        except (TypeError, ValueError):
            print('Incorrect value - enter again...')


def parse_value(raw_string):
    """Parse the entered value with units"""
    factor = 1
    string = raw_string.strip()
    for symbol in Measure.symbols:
        # End after first match
        if string.endswith(symbol):
            string = string.replace(symbol, '')
            # Default unit - 1pt
            factor = Measure.units.get(symbol, 1)
            break
    else:
        factor = Measure.units.get(Measure.default_unit, 1)
    # Filter the string
    num_string = ''.join(x for x in string if x.isdigit() or x is '.')
    # Convert the value with known unit to DTP points
    points = float(num_string) * factor
    return round(points, 2)



# -*- coding: utf-8 -*-
"""User help messages for rpi2caster"""


casting_menu = (
    """This program lets you:

    -cast composition or punch paper ribbon
        The program asks you how many times you want to cast the job,
        how many lines to skip (for casting multi-line jobs),
        then casts type or punches ribbon, displaying information about
        current combination, progress, wedge positions and pump status.
        After the last run is finished, the program will pause and ask you
        whether you want to cast it once more. Otherwise, you return to menu.

        Sometimes the machine stops or you need to abort casting type;
        for an emergency stop, press Ctrl-C. The program will send signals
        to stop the pump (if it was working), and ask you whether you want
        to continue or abort. You can retry the last run; in this case,
        you can skip all lines cast successfully. Two lines will be cast
        to heat up the mould, so that the temperature is stable.

    -cast sorts from specified matrix coordinates, e.g. H2
        Enter a number of units to which the character will be cast.
        Enter a number of characters.
        The program will let you enter as many as you want. After you decide
        to cast the queue, the program asks for the desired line length
        (set the galley width accordingly), calculates the number of sorts
        to fits in a line. Characters are separated with spaces to prevent
        matrix overheating. The program will cast one or more lines,
        the remaining part of the line will be quadded out.

    -cast typecases from diecase for a given language
        This option is available after you choose a diecase.
        Choose a language

    -add, edit and delete diecases

    -do caster/interface diagnostics and calibration
    """)



# -*- coding: utf-8 -*-
"""Module containing functions for diecase retrieval and parsing,
diecase storing, parameter searches.
"""
# File operations
import io
import csv
from copy import copy
# Some functions raise custom exceptions
from . import exceptions as e
# Constants module
from . import constants as c
# Parsing module
from . import parsing as p
# Styles manager
from .styles import Styles
# Wedge operations for several matrix-case management functions
from .wedge_data import Wedge
# User interface
from .global_config import UI, APPDIR
# Letter frequency for completeness testing
from .letter_frequencies import CharFreqs
# Database backend
from .database import Database
# Make or use existing database instance
DB = Database()


class Diecase(object):
    """Diecase: matrix case attributes and operations"""
    def __init__(self, diecase_id='', manual_choice=False):
        if manual_choice:
            data = choose_diecase()
        elif diecase_id:
            try:
                data = DB.get_diecase(diecase_id)
            except (e.NoMatchingData, e.DatabaseQueryError):
                data = choose_diecase()
        else:
            data = None
        if not data:
            # Use default diecase
            UI.debug_info('Using empty 15x17 diecase with S5-12E wedge...')
            data = ['', '', 'S5-12E', []]
        UI.debug_info('Processing diecase data...')
        (self.diecase_id, self.typeface, wedge_name, layout) = data
        self.wedge = Wedge(wedge_name, manual_choice=not wedge_name)
        self.alt_wedge = self.wedge
        self.layout = layout

    def __iter__(self):
        return iter(self.matrices)

    def __repr__(self):
        return self.diecase_id

    def __bool__(self):
        return bool(self.diecase_id)

    def __getitem__(self, item):
        try:
            char, styles = item
        except ValueError:
            char = item
            styles = ''
        return self.lookup_matrix(char, styles)

    @property
    def matrices(self):
        """Gets an iterator of mats"""
        matrices = self.__dict__.get('_matrices')
        if not matrices:
            layout = (self.__dict__.get('_layout') or
                      generate_empty_layout(15, 17))
            mats = [Matrix(char=char, styles=styles, code=code,
                           units=units, diecase=self)
                    for (char, styles, code, units) in layout]
            self.__dict__['_matrices'] = mats
        return (x for x in self.__dict__.get('_matrices'))

    @matrices.setter
    def matrices(self, matrices):
        """Sets a list of matrices"""
        self.__dict__['_matrices'] = [mat for mat in matrices]

    @property
    def ligatures(self):
        """Matrix character may be any string, not just a single character.
        This is used for ligatures, like ct, fb, fh, fi, fl, ffi, ffl, ij etc.
        Return an iterator of matrices where char attribute is not a single
        character. En-dashes and em-dashes will end up here also."""
        return (x for x in self.matrices if len(x.char) > 1)

    @property
    def ligature_length(self):
        """Get the maximum ligature length in chars"""
        return max(len(matrix) for matrix in self.matrices)

    @property
    def low_spaces(self):
        """Return an iterator of low spaces"""
        return (x for x in self.matrices if x.islowspace())

    @property
    def high_spaces(self):
        """Return an iterator of high spaces"""
        return (x for x in self.matrices if x.ishighspace())

    @property
    def spaces(self):
        """Return an iterator of all spaces, low and high"""
        return (x for x in self.matrices if x.isspace())

    @property
    def layout(self):
        """Gets a diecase layout as a list of lists"""
        return [matrix.get_record() for matrix in self.matrices]

    @layout.setter
    def layout(self, layout):
        """Translates the layout to a list of matrix objects"""
        if layout:
            self.matrices = []
            self.__dict__['_layout'] = layout

    @property
    def parameters(self):
        """Gets a list of parameters"""
        return [(self.diecase_id, 'Diecase ID'),
                (self.typeface, 'Typeface'),
                (self.wedge, 'Assigned wedge')]

    @property
    def styles(self):
        """Parses the diecase layout and gets available typeface styles.
        Returns a list of them."""
        all_styles = set(''.join([mat.styles for mat in self]))
        return Styles(all_styles)()

    def show_layout(self):
        """Shows the diecase layout"""
        UI.display_diecase_layout(self)
        UI.pause()

    def edit_layout(self):
        """Edits a layout and asks if user wants to save changes"""
        # Edit the layout
        UI.edit_diecase_layout(self)

    def lookup_matrix(self, char='', style=''):
        """Chooses a matrix automatically or manually (if multiple matches),
        allows to specify matrix data manually if no matches found"""
        def enter_manually(char='', style=''):
            """Specify matrix by entering code"""
            if char:
                st_string = str(Styles(style))
                if 'all' in st_string:
                    st_string = ''
                else:
                    st_string = st_string + ' '
                what = spaces.get(char) or '%s%s' % (st_string, char)
                UI.display('Choose matrix for %s' % what)
            else:
                char = ''
            code = UI.enter_data_or_default('Mat position?', 'G5').upper()
            matrix = Matrix(char=char, style=style, code=code, diecase=self)
            return matrix

        spaces = {' ': 'low space', '_': 'high space'}
        if char in spaces:
            style = ''
        # Find as many matches as we can:
        # -all matrices for a given character for a given style
        # -all matrices for a given style if char not specified
        # -all matrices for a given character if style not specified
        # -all matrices for spaces
        # -enter manually if no character and no style is specified
        # Then choose from menu (if multiple match), automatically
        # (single match) or enter manually (no matches)
        if char and style:
            candidates = [mat for mat in self.matrices
                          if mat.char == char and style in mat.styles]
        elif char:
            candidates = [mat for mat in self.matrices if mat.char == char]
        elif style:
            candidates = [mat for mat in self.matrices
                          if style in mat.styles or not self.styles]
        else:
            candidates = []
        # Built the list of candidates...
        if not candidates:
            return enter_manually(char, style)
        elif len(candidates) == 1:
            return candidates[0]
        else:
            # Multiple matches found = let user choose
            st_string = str(Styles(style))
            if 'all' in st_string:
                st_string = ''
            else:
                st_string = ' ' + st_string
            what = spaces.get(char) or '%s%s' % (st_string, char)
            UI.display_header('Multiple matrices for %s' % what)
            # Show a menu with multiple candidates
            mats = {i: mat for i, mat in enumerate(candidates, start=1)}
            UI.display(''.join(['Index'.ljust(10), 'Char'.ljust(10),
                                'Styles'.ljust(30), 'Coordinates']))
            for i, mat in mats.items():
                matrix_styles = str(Styles(mat.styles))
                record = [str(i).ljust(10), mat.char.ljust(10),
                          matrix_styles.ljust(30), mat.code]
                UI.display(''.join(record))
            prompt = 'Choose matrix (leave blank to enter manually)'
            choice = UI.enter_data_or_blank(prompt, int)
            matrix = mats.get(choice) or enter_manually(char, style)
            return matrix

    def decode_matrix(self, code):
        """Finds the matrix based on the column and row in layout"""
        matrices = (mat for mat in self.matrices if mat.code == code.upper())
        try:
            mat = next(matrices)
        except StopIteration:
            mat = Matrix(code=code, diecase=self)
        if not mat.char:
            # Assume it's a space
            mat.char = ' '
        return mat

    def import_layout(self):
        """Imports a layout from file"""
        # Load the layout from file
        submitted_layout = [x for x in import_layout_file()]
        if not submitted_layout:
            UI.pause('File does not contain a proper layout!')
            return False
        # Update the empty layout with characters read from file
        # record = (char, styles, coordinates, units)
        try:
            self.layout = generate_empty_layout()
            for matrix in self.matrices:
                for record in submitted_layout:
                    (char, styles, coordinates, units) = record
                    if matrix.code.upper() == coordinates.upper():
                        matrix.char = char
                        matrix.styles = styles
                        matrix.units = units
        except (TypeError, ValueError):
            UI.pause('Cannot process the uploaded layout!')
            return False

    def export_layout(self):
        """Exports the matrix case layout to file."""
        name = self.diecase_id or 'NewDiecase'
        # Save the exported diecase layout in the default directory
        filename = APPDIR + '/%s.csv' % name
        with io.open(filename, 'a') as output_file:
            csv_writer = csv.writer(output_file, delimiter=';', quotechar='"',
                                    quoting=csv.QUOTE_ALL)
            csv_writer.writerow(['Char', 'Styles', 'Coordinates', 'Units'])
            for matrix in self.matrices:
                csv_writer.writerow(matrix.get_record())
        UI.pause('File %s successfully saved.' % filename)
        return True

    def test_characters(self, input_string='', styles='r'):
        """Enter the string and parse the diecase to see if any of the
        specified characters are missing."""
        input_string = input_string or UI.enter_data('Text to check?')
        charset = {char for char in input_string}
        # Return True unless any characters are missing
        retval = True
        for style, description in Styles(styles).items():
            chars_found = {mat.char for mat in self for char in charset
                           if char == mat.char and
                           (mat.isspace() or style in mat.styles)}
            missing = sorted(charset.difference(chars_found))
            if missing:
                UI.display('Missing mats for %s: %s'
                           % (description, ', '.join(missing)))
                retval = False
            else:
                UI.display('All characters for %s are present.' % description)
        return retval

    def _test_lang_completeness(self):
        """Choose a language and test whether the diecase contains all
        characters, and if not, list them"""
        char_backend = CharFreqs()
        UI.display('Building character set...')
        uppercase = [char.upper() for char in char_backend]
        lowercase = [char.lower() for char in char_backend]
        all_chars = ''.join(list(sorted(set(uppercase + lowercase))))
        UI.display('\nCharacters: %s\n' % all_chars)
        styles = Styles(self.styles, manual_choice=True)()
        self.test_characters(all_chars, styles)
        UI.pause()

    def _set_diecase_id(self, diecase_id=None):
        """Sets a diecase ID"""
        prompt = 'Diecase ID? (leave blank to exit)'
        diecase_id = (diecase_id or UI.enter_data_or_blank(prompt) or
                      self.diecase_id)
        # Ask if we are sure we want to update this
        # if self.diecase_id was set earlier
        if not self.diecase_id or UI.confirm('Apply changes?', default=False):
            self.diecase_id = diecase_id
            return True

    def _set_typeface(self, typeface=None):
        """Sets the type series, size and typeface name"""
        prompt = 'Typeface (series, size, name)?'
        typeface = typeface or UI.enter_data_or_blank(prompt) or self.typeface
        if not self.typeface or UI.confirm('Apply changes?', default=False):
            self.typeface = typeface
            return True

    def _assign_wedge(self, wedge=''):
        """Assigns a wedge (from database or newly-defined) to the diecase"""
        self.wedge = Wedge(wedge, manual_choice=not wedge)

    def _save_to_db(self):
        """Stores the matrix case definition/layout in database"""
        try:
            DB.add_diecase(self)
            UI.pause('Data saved successfully.')
            return True
        except e.DatabaseQueryError:
            UI.pause('Cannot save the diecase!')

    def _delete_from_db(self):
        """Deletes a diecase from database"""
        ans = UI.confirm('Are you sure?', default=False)
        if ans and DB.delete_diecase(self):
            UI.display('Matrix case deleted successfully.')
            return True

    def _check_db(self):
        """Checks if the diecase is registered in database"""
        try:
            DB.get_diecase(self.diecase_id)
            return True
        except (e.DatabaseQueryError, e.NoMatchingData):
            return False

    def manipulation_menu(self):
        """A menu with all operations on a diecase"""
        def clear_layout():
            """Generates a new layout for the diecase"""
            self.layout = generate_empty_layout()
        try:
            while True:
                UI.clear()
                UI.display_parameters({'Diecase data': self.parameters,
                                       'Wedge data': self.wedge.parameters})
                options = {'Q': e.return_to_menu,
                           'T': self._set_typeface,
                           'A': self._test_lang_completeness,
                           'W': self._assign_wedge,
                           'ID': self._set_diecase_id,
                           'E': self.edit_layout,
                           'N': clear_layout,
                           'V': self.show_layout,
                           'I': self.import_layout,
                           'X': self.export_layout,
                           'C': e.menu_level_up}
                messages = ['Matrix case manipulation:\n\n'
                            '[V]iew layout,\n'
                            '[E]dit layout,\n'
                            '[I]mport layout from file\n'
                            'e[X]port layout to file,\n'
                            'assign [W]edge,\n'
                            'change [T]ypeface\n'
                            'set diecase [ID]\n'
                            'check if [A]ll characters for chosen language '
                            'are in diecase,\n'
                            '[N]ew layout from scratch\n']
                # Save to database needs a complete set of metadata
                required = {'Typeface': self.typeface,
                            'Diecase ID': self.diecase_id}
                missing = [item for item in required if not required[item]]
                messages.extend(['Cannot save diecase with missing %s\n'
                                 % item.lower() for item in missing])
                if not missing:
                    options['S'] = self._save_to_db
                    messages.append('[S]ave diecase to database,\n')
                # Check if it's in the database
                if self._check_db():
                    options['D'] = self._delete_from_db
                    messages.append('[D]elete diecase from database,\n')
                # Options constructed
                messages.append('[C] to choose/create another diecase,\n'
                                '[Q] to quit.\n')
                messages.append('Your choice: ')
                message = ''.join(messages)
                UI.simple_menu(message, options)()
        except e.MenuLevelUp:
            # Exit matrix case manipulation menu
            return True


class Matrix(object):
    """A class for single matrices - all matrix data"""
    def __init__(self, **kwargs):
        self.diecase = kwargs.get('diecase', EMPTY_DIECASE)
        self.char = kwargs.get('char', '')
        self.styles = kwargs.get('styles', 'r')
        self.code = kwargs.get('code', 'O15')
        self.units = kwargs.get('units') or 0

    def __repr__(self):
        return self.code

    def __str__(self):
        if self.points != self.get_row_points():
            s_signal = 'S'
        else:
            s_signal = ''
        comment = (self.islowspace() and
                   ' // low space %.2f points wide' % self.points or
                   self.ishighspace() and
                   ' // high space %.2f points wide' % self.points or
                   self.char and
                   ' // %s %s' % (Styles(self.styles), self.char) or '')
        return '%s%s%s%s' % (self.column, s_signal, self.row, comment)

    def __call__(self, points):
        """Returns a copy of self with a corrected width"""
        new_mat = copy(self)
        new_mat.points += points
        return new_mat

    def __len__(self):
        return len(self.char)

    def __bool__(self):
        return True

    @property
    def styles(self):
        """Get the matrix's styles or an empty string if matrix has no char"""
        if not self.char or self.isspace():
            return ''
        else:
            return self.__dict__.get('_styles', 'r')

    @styles.setter
    def styles(self, styles):
        """Sets the matrix's style string"""
        self.__dict__['_styles'] = styles

    @property
    def units(self):
        """Gets the specific or default number of units"""
        # If units are assigned to the matrix, return the value
        # Otherwise, wedge default
        units = (self.__dict__.get('_units', 0) or
                 self.diecase.wedge.units[self.row])
        return units

    @units.setter
    def units(self, units):
        """Sets the unit width value"""
        if units:
            self.__dict__['_units'] = units

    @property
    def points(self):
        """Gets a DTP point (=.1667"/12) width for the matrix"""
        points = (self.__dict__.get('_points') or
                  self.diecase.wedge.units_to_points_factor * self.units)
        return round(points, 2)

    @points.setter
    def points(self, points):
        """Sets a DTP point value for the mat and updates the unit value"""
        # Set constraints on width
        lim_points = min(points, self.get_max_points())
        lim_points = max(lim_points, self.get_min_points())
        self.__dict__['_points'] = lim_points

    @property
    def row(self):
        """Gets the row number"""
        return self.__dict__.get('_row', 15)

    @row.setter
    def row(self, value):
        """Sets the row number"""
        value = int(value)
        if value in range(1, 17):
            self.__dict__['_row'] = value

    @property
    def column(self):
        """Gets the matrix column"""
        return self.__dict__.get('_column', 'O')

    @column.setter
    def column(self, value):
        """Sets the column"""
        column = value.upper()
        if column in c.COLUMNS_17:
            self.__dict__['_column'] = column
        else:
            self.__dict__['_column'] = 'O'

    @property
    def code(self):
        """Gets the matrix code"""
        return '%s%s' % (self.column, self.row)

    @code.setter
    def code(self, code_string):
        """Sets the coordinates for the matrix"""
        signals = p.parse_signals(code_string)
        self.row = p.get_row(signals)
        self.column = p.get_column(signals)

    def get_row_points(self):
        """Gets a number of points for characters in the diecase row"""
        # Try wedges in order:
        # diecase's temporary wedge, diecase's default wedge, standard S5-12E
        return self.diecase.alt_wedge.points[self.row]

    def get_min_points(self):
        """Gets the minimum unit value for a given wedge, based on the
        matrix row and wedge unit value, for wedges at 1/1"""
        # 3/8 - 1/1 = 2/7 so 2*15 + 7 = 37
        # 37 * 0.0005 = 0.0185 i.e. max inches we can take away
        return round(max(self.get_row_points() - 0.0185 * 72, 0), 2)

    def get_max_points(self):
        """Gets the minimum unit value for a given wedge, based on the
        matrix row and wedge unit value, for wedges at 1/1"""
        # 15/15 - 3/8 = 12/7 so 12*15 + 7 = 187
        # 187 * 0.0005 = 0.0935 i.e. max inches we can add
        # Matrix is typically .2 x .2 - anything wider will lead to a splash!
        # Safe limit is .19" = 13.680" and we don't let the mould open further
        full_width = round(self.get_row_points() + 0.0935 * 72, 2)
        width_cap = 13.680
        if self.islowspace():
            return full_width
        else:
            return min(full_width, width_cap)

    def wedge_positions(self, point_correction=0):
        """Calculate the 0075 and 0005 wedge positions for this matrix
        based on the current wedge used"""
        diff = self.points + point_correction - self.get_row_points()
        # The following calculations are done in 0005 wedge steps
        # 1 step of 0075 wedge is 15 steps of 0005; neutral positions are 3/8
        # 3 * 15 + 8 = 53, so any increment/decrement is relative to this
        # Add or take away a number of inches; diff is in points i.e. 1/72"
        steps_0005 = int(diff * 2000 / 72) + 53
        # Upper limit: 15/15 => 15*15=225 + 15 = 240
        # Unsafe for casting from mats - .2x.2 size - larger leads to splash!
        # Adding a high space for overhanging character is the way to do it
        # Space casting is fine, we can open the mould as far as possible
        steps_0005 = min(steps_0005, 240)
        # Lower limit: 1/1 wedge positions => 15 + 1 = 16:
        steps_0005 = max(16, steps_0005)
        steps_0075 = 0
        while steps_0005 > 15:
            steps_0005 -= 15
            steps_0075 += 1
        # Got the wedge positions, return them
        return (steps_0075, steps_0005)

    def get_record(self):
        """Returns a record suitable for JSON-dumping and storing in DB"""
        return [self.char, self.styles, self.code, self.units]

    def isspace(self):
        """Check whether the mat is space or not"""
        return self.islowspace() or self.ishighspace()

    def islowspace(self):
        """Check whether the mat is low space"""
        return self.char == ' '

    def ishighspace(self):
        """Check whether the mat is high space"""
        return self.char == '_'

    def edit(self):
        """Edits the matrix data"""
        char = self.char and 'char: "%s"' % self.char or 'character not set'
        styles = self.styles and 'styles: %s' % self.styles or ''
        UI.display('\n' + '*' * 80 + '\n')
        UI.display('%s %s %s, units: %s'
                   % (self.code, styles, char, self.units))
        self.edit_char()
        self.choose_styles()
        self.specify_units()
        return self

    def edit_char(self):
        """Edit the matrix character"""
        prompt = ('Char? (" ": low / "_": high space, '
                  'blank = keep, ctrl-C to exit)?')
        if self.char:
            self.char = UI.enter_data_or_default(prompt, self.char)
        else:
            self.char = UI.enter_data_or_blank(prompt)

    def specify_units(self):
        """Give a user-defined unit value for the matrix;
        mostly used for matrices placed in a different row than the
        unit arrangement indicates; this will make the program set the
        justification wedges and cast the character with the S-needle"""
        UI.display('Enter unit value for %s; 0 = row units, '
                   'blank = current' % self.code)
        prompt = 'Units?'
        self.units = UI.enter_data_or_default(prompt, self.units, int)

    def choose_styles(self):
        """Choose styles for the matrix"""
        style_manager = Styles(self.styles)
        style_manager.choose()
        self.styles = style_manager()


def diecase_operations():
    """Matrix case operations menu for inventory management"""
    try:
        UI.display_header('Matrix case manipulation')
        while True:
            # Choose a diecase or initialize a new one
            Diecase(manual_choice=True).manipulation_menu()
    except e.ReturnToMenu:
        # Exit diecase operations
        return True


def list_diecases():
    """Lists all matrix cases we have."""
    data = DB.get_all_diecases()
    results = {}
    UI.display('\n' +
               'No.'.ljust(4) +
               'Diecase ID'.ljust(25) +
               'Wedge'.ljust(12) +
               'Typeface' +
               '\n\n0 - start a new empty diecase\n')
    for index, diecase in enumerate(data, start=1):
        # Start each row with index
        row = [str(index).ljust(4)]
        # Collect the ciecase parameters: ID, typeface, wedge
        # Leave the diecase layout out
        row.append(diecase[0].ljust(25))
        # Swap the wedge and typeface designations (more place for long names)
        row.append(diecase[2].ljust(12))
        row.append(diecase[1])
        # Add number and ID to the result that will be returned
        results[index] = diecase[0]
        UI.display(''.join(row))
    UI.display('\n\n')
    # Now we can return the number - diecase ID pairs
    return results


def choose_diecase():
    """Lists diecases and lets the user choose one;
    returns the Diecase class object with all parameters set up."""
    prompt = 'Number of a diecase (0 for a new one, leave blank to exit)'
    while True:
        try:
            UI.display('Choose a matrix case:', end='\n\n')
            data = list_diecases()
            choice = UI.enter_data_or_exception(prompt, e.ReturnToMenu, int)
            if not choice:
                return None
            else:
                return DB.get_diecase(data[choice])
        except KeyError:
            UI.pause('Diecase number is incorrect!')
        except (e.NoMatchingData, e.DatabaseQueryError):
            UI.display('No diecases found in database')
            return None


def import_layout_file():
    """Reads a matrix case arrangement from a text or csv file.
    The format should be: "character";"styles";"coordinates";"unit_width"."""
    # Give us a file or end here
    try:
        filename = UI.enter_input_filename()
    except e.ReturnToMenu:
        return False
    # Initialize the records list
    raw_records = []
    # This will store the processed combinations - and whenever a duplicate
    # is detected, the function will raise an exception
    with io.open(filename, 'r') as layout_file:
        input_data = csv.reader(layout_file, delimiter=';', quotechar='"')
        raw_records = [record for record in input_data]
        displayed_lines = [' '.join(record) for record in raw_records[:5]]
        # Preview file
        UI.display('File preview: displaying first 5 rows:\n')
        UI.display('\n'.join(displayed_lines), end='\n\n')
        # Ask if the first row is a header - if so, away with it
        if UI.confirm('Is the 1st row a table header? ', default=True):
            raw_records.pop(0)
    if not UI.confirm('Proceed?', default=True):
        return False
    try:
        # Assume we have a 15x15 diecase (if it turns out larger - override)
        columns = c.COLUMNS_15
        rows = range(1, 16)
        records = []
        for record in raw_records:
            # Must work also if we don't specify the units
            try:
                (char, styles, coordinates, units) = record
                units = int(units.strip())
            except (ValueError, AttributeError):
                (char, styles, coordinates) = record[:3]
                units = 0
            # Strip unneeded whitespace from character
            if char.isspace():
                char = ' '
            else:
                char = char.strip()
            # Parse styles and order them
            styles = Styles(styles)()
            # Determine the diecase size, override previous size if larger
            if '16' in coordinates:
                columns = c.COLUMNS_17
                rows = range(1, 17)
            elif 'NI' in coordinates or 'NL' in coordinates:
                columns = c.COLUMNS_17
            records.append((char, styles, coordinates, units))
        # Arrange the uploaded list into a layout
        layout = (record for row in rows for col in columns
                  for record in records if record[2] == '%s%s' % (col, row))
        # Show the uploaded layout
        return layout
    except (KeyError, ValueError, IndexError):
        return []


def generate_empty_layout(rows=None, columns=None):
    """Makes a list of empty matrices, row for row, column for column"""
    prompt = ('Matrix case size: 1 for 15x15, 2 for 15x17, 3 for 16x17? '
              '(leave blank to cancel) : ')
    options = {'1': (15, 15), '2': (15, 17), '3': (16, 17), '': ()}
    if (rows, columns) not in options.values():
        choice = UI.simple_menu(prompt, options)
        if not choice:
            return
        (rows, columns) = choice
    # Generate column numbers
    columns_list = columns == 17 and c.COLUMNS_17 or c.COLUMNS_15
    # Generate row numbers: 1...15 or 1...16
    rows_list = [num + 1 for num in range(rows)]
    return (('', '', '%s%s' % (column, row), 0)
            for row in rows_list for column in columns_list)


# Make a single instance of empty diecase for getting fallback values
EMPTY_DIECASE = Diecase()



# -*- coding: utf-8 -*-
"""Operations on ribbon and scheme objects: creating, editing and deleting. """
# File operations
import io
# Some functions raise custom exceptions
from . import exceptions as e
# Constants for rpi2caster
from .constants import ASSIGNMENT_SYMBOLS
# User interface
from .global_config import UI
# Database
from . import database
DB = database.Database()


class Ribbon(object):
    """Ribbon objects - no matter whether files or database entries.

    A ribbon has the following attributes:
    description, customer - strings
    diecase_id (diecase is selected automatically on casting, so user can e.g.
                view the layout or know which wedge to use)
    contents - series of Monotype codes

    Methods:
    choose_ribbon - choose ribbon automatically or manually,
        first try to get a ribbon with ribbon_id, and if that fails
        ask and select ribbon manually from database, and if that fails
        ask and load ribbon from file
    read_from_file - select a file, parse the metadata, set the attributes
    get_from_db - get data from database
    export_to_file - store the metadata and contents in text file
    store_in_db - store the metadata and contents in db
    set_[description, customer, diecase_id] - set parameters manually"""
    def __init__(self, ribbon_id='', filename='', manual_choice=False):
        data = None
        self.filename = ''
        if ribbon_id:
            try:
                data = DB.get_ribbon(ribbon_id)
            except (e.NoMatchingData, e.DatabaseQueryError):
                UI.display('Ribbon choice failed. Starting a new one.')
        elif filename:
            self.filename = filename
            data = import_ribbon_from_file(filename)
        elif manual_choice:
            data = choose_ribbon_from_db() or import_ribbon_from_file()
        # Got data - unpack them, set the attributes
        if not data:
            data = ['', 'New ribbon', 'No customer', '', 'S5-12E', []]
        (self.ribbon_id, self.description, self.customer, self.diecase_id,
         self.wedge_name, self.contents) = data

    def __iter__(self):
        return iter(self.contents)

    def __next__(self):
        yield from self.contents

    def __repr__(self):
        return self.ribbon_id or ''

    def __bool__(self):
        return bool(self.contents)

    def set_ribbon_id(self, ribbon_id=None):
        """Sets the ribbon ID"""
        prompt = 'Ribbon ID? (leave blank to exit) : '
        ribbon_id = (ribbon_id or UI.enter_data_or_blank(prompt) or
                     self.ribbon_id)
        # Ask if we are sure we want to update this
        # if self.ribbon_id was set earlier
        prompt = 'Are you sure to change the ribbon ID?'
        condition = not self.ribbon_id or UI.confirm(prompt, default=False)
        if condition:
            self.ribbon_id = ribbon_id
            return True

    def set_description(self, description=None):
        """Manually sets the ribbon's description"""
        prompt = 'Enter the title: '
        self.description = (description or
                            UI.enter_data_or_blank(prompt) or self.description)

    def set_customer(self, customer=None):
        """Manually sets the customer"""
        prompt = 'Enter the customer\'s name for this ribbon: '
        self.customer = (customer or
                         UI.enter_data_or_blank(prompt) or self.customer)

    def set_diecase(self, diecase_id=''):
        """Chooses the diecase for this ribbon"""
        self.diecase_id = diecase_id

    def set_wedge(self, wedge_name=None):
        """Sets a wedge for the ribbon"""
        self.wedge_name = wedge_name

    @property
    def parameters(self):
        """Gets a list of parameters"""
        return [(self.filename, 'File name'),
                (self.ribbon_id, 'Ribbon ID'),
                (self.description, 'Description'),
                (self.customer, 'Customer'),
                (self.diecase_id, 'Matrix case ID'),
                (self.wedge_name, 'Wedge')]

    def display_contents(self):
        """Displays the ribbon's contents, line after line"""
        UI.display('Ribbon contents preview:\n')
        contents_generator = (line for line in self.contents if line)
        try:
            while True:
                UI.display(contents_generator.__next__())
        except StopIteration:
            # End of generator
            UI.pause('Finished', UI.MSG_MENU)
        except (EOFError, KeyboardInterrupt):
            # Press ctrl-C to abort displaying long ribbons
            UI.pause('Aborted', UI.MSG_MENU)

    def get_from_db(self):
        """Gets the ribbon from database"""
        data = choose_ribbon_from_db()
        if data and UI.confirm('Override current data?', default=False):
            (self.ribbon_id, self.description, self.customer, self.diecase_id,
             self.wedge_name, self.contents) = data

    def store_in_db(self):
        """Stores the ribbon in database"""
        UI.display_parameters({'Ribbon data': self.parameters})
        # Ask for confirmation
        try:
            DB.add_ribbon(self)
            UI.pause('Ribbon added successfully.')
            return True
        except (e.DatabaseQueryError, e.NoMatchingData):
            UI.pause('Cannot store ribbon in database!')
            return False

    def delete_from_db(self):
        """Deletes a ribbon from database."""
        if UI.confirm('Are you sure?', default=False):
            DB.delete_ribbon(self)
            UI.display('Ribbon deleted successfully.')

    def import_from_file(self, filename=None):
        """Reads a ribbon file, parses its contents, sets the ribbon attrs"""
        self.filename = filename
        (self.ribbon_id, self.description, self.customer, self.diecase_id,
         self.wedge_name, self.contents) = import_ribbon_from_file()

    def export_to_file(self, filename=None):
        """Exports the ribbon to a text file"""
        UI.display_parameters({'Ribbon data': self.parameters})
        try:
            # Choose file, write metadata, write contents
            filename = filename or UI.enter_output_filename()
        except e.ReturnToMenu:
            return
        with io.open(filename, mode='w+') as ribbon_file:
            ribbon_file.write('description: ' + self.description)
            ribbon_file.write('customer: ' + self.customer)
            ribbon_file.write('diecase: ' + self.diecase_id)
            ribbon_file.write('wedge: ' + self.wedge_name)
            for line in self.contents:
                ribbon_file.write(line)


def list_ribbons():
    """Lists all ribbons in the database."""
    data = DB.get_all_ribbons()
    results = {}
    UI.display('\n' +
               'No.'.ljust(5) +
               'Ribbon name'.ljust(30) +
               'Description'.ljust(30) +
               'Customer'.ljust(30) +
               'Diecase'.ljust(30) +
               'Wedge'.ljust(30) +
               '\n\n0 - start a new empty ribbon\n')
    for index, ribbon in enumerate(data, start=1):
        # Collect ribbon parameters
        row = [str(index).ljust(5)]
        # Add ribbon name, description, diecase ID
        row.extend([field.ljust(30) for field in ribbon[:-1]])
        # Display it all
        UI.display(''.join(row))
        # Add number and ID to the result that will be returned
        results[index] = ribbon[0]
    UI.display('\n\n')
    # Now we can return the number - ribbon ID pairs
    return results


def import_ribbon_from_file(filename=None):
    """Imports ribbon from file, parses parameters, returns a list of them"""
    def get_value(line, symbol):
        """Helper function - strips whitespace and symbols"""
        # Split the line on an assignment symbol, get the second part,
        # strip any whitespace or multipled symbols
        return line.split(symbol, 1)[-1].strip(symbol).strip()

    UI.display('Loading the ribbon from file...')
    try:
        filename = filename or UI.enter_input_filename()
        with io.open(filename, mode='r') as ribbon_file:
            ribbon = [line.strip() for line in ribbon_file if line.strip()]
    except (FileNotFoundError, IOError):
        UI.pause('Cannot open ribbon file %s' % filename)
        return False
    except e.ReturnToMenu:
        return False
    # What to look for
    keywords = ['diecase', 'description', 'desc', 'diecase_id', 'customer',
                'wedge', 'stopbar']
    # Metadata (anything found), contents (the rest)
    metadata = {}
    contents = []
    # Look for parameters line per line, get parameter value
    # If parameters exhausted, append the line to contents
    for line in ribbon:
        for keyword in keywords:
            if line.startswith(keyword):
                for sym in ASSIGNMENT_SYMBOLS:
                    if sym in line:
                        # Data found
                        metadata[keyword] = get_value(line, sym)
                        break
                break
        else:
            contents.append(line)
    # Metadata parsing
    ribbon_id = None
    description = metadata.get('description', '') or metadata.get('desc', '')
    customer = metadata.get('customer', '')
    diecase_id = metadata.get('diecase', '') or metadata.get('diecase_id', '')
    wedge = metadata.get('wedge', '') or metadata.get('stopbar', '')
    # Add the whole contents as the attribute
    return (ribbon_id, description, customer, diecase_id, wedge, contents)


def choose_ribbon_from_db():
    """Chooses ribbon data from database"""
    prompt = 'Number of a ribbon? (0 for a new one, leave blank to exit): '
    while True:
        try:
            data = list_ribbons()
            choice = UI.enter_data_or_exception(prompt, e.ReturnToMenu, int)
            if not choice:
                return None
            else:
                return DB.get_ribbon(data[choice])
        except KeyError:
            UI.pause('Ribbon number is incorrect!')
        except (e.DatabaseQueryError, e.NoMatchingData):
            UI.display('No ribbons found in database')
            return None



"""texlib.wrap

Implements TeX's algorithm for breaking paragraphs into lines.

This module provides a straightforward implementation of the algorithm
used by TeX to format paragraphs.  The algorithm uses dynamic
programming to find a globally optimal division into lines, enabling
it to produce more attractive results than a first-fit or best-fit
algorithm can.  For a full description, see the reference.

The module provides the ObjectList class, which is a list of Box,
Glue, and Penalty instances.  The elements making up a paragraph of
text should be assembled into a single ObjectList.  Boxes represent
characters of type, and their only attribute is width.  Glue
represents a space of variable size; in addition to a preferred width,
glue can also stretch and shrink, to an amount that's specified by the
user.  Penalties are used to encourage or discourage breaking a line
at a given point.  Positive values discourage line breaks at a given
point, and a value of INFINITY forbids breaking the line at the
penalty.  Negative penalty values encourage line breaks at a given
point, and a value of -INFINITY forces a line break at a particular
point.

The compute_breakpoints() method of ObjectList returns a list of
integers containing the indexes at which the paragraph should be
broken.  If you're setting the text to be ragged-right (or
ragged-left, I suppose), then simply loop over the text and insert
breaks at the appropriate points.  For full justification, you'll have
to loop over each line's contents, calculate its adjustment ratio by
calling compute_adjustment_ratio(), and for each bit of glue, call its
compute_width() method to figure out how long this dab of glue should
be.

Reference:
    "Breaking Paragraphs into Lines", D.E. Knuth and M.F. Plass,
    chapter 3 of _Digital Typography_, CSLI Lecture Notes #78.

This software is available under the MIT license.

Copyright (c) 2010, A.M. Kuchling

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included
  in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import sys

INFINITY = 1000

# Three classes defining the three different types of object that
# can go into an ObjectList.


class Box:
    """Class representing a glyph or character.  Boxes have a fixed
    width that doesn't change.
    """

    def __init__(self, b_width, character=None):
        self.character = character
        self.width = b_width
        self.stretch = 0
        self.shrink = 0
        self.penalty = 0
        self.flagged = 0
        self.is_glue = False
        self.is_box = True
        self.is_penalty = False
        self.is_forced_break = False


class Glue:
    """Class representing a bit of glue.  Glue has a preferred width,
    but it can stretch up to an additional distance, and can shrink
    by a certain amount.  Line breaks can be placed at any point where
    glue immediately follows a box.
    """

    def __init__(self, width, shrink, stretch):
        self.width, self.stretch, self.shrink = width, stretch, shrink
        self.is_glue = True
        self.is_box = False
        self.is_penalty = False
        self.is_forced_break = False

    def compute_width(self, ratio):
        """Return how long this glue should be, for the given adjustment
        ratio r."""
        if ratio < 0:
            return self.width + ratio * self.shrink
        else:
            return self.width + ratio * self.stretch


class Penalty:
    """Class representing a penalty.  Negative penalty values
    encourage line breaks at a given point, and positive values
    discourage breaks.  A value of INFINITY either absolutely requires
    or forbids a break.  Penalties have a width of zero unless a break
    is taken at the penalty point, at which point the value of the
    penalty's 'width' attribute is used.

    """

    def __init__(self, width, penalty, flagged=0):
        self.width = width
        self.penalty = penalty
        self.flagged = flagged
        self.stretch = self.shrink = 0
        self.is_glue = False
        self.is_box = False
        self.is_penalty = True
        self.is_forced_break = self.penalty == INFINITY


class _BreakNode:
    "Internal class representing an active breakpoint."

    def __init__(self, position, line, fitness_class,
                 totalwidth, totalstretch, totalshrink,
                 demerits, previous=None):
        self.position, self.line = position, line
        self.fitness_class = fitness_class
        self.totalwidth, self.totalstretch = totalwidth, totalstretch
        self.totalshrink, self.demerits = totalshrink, demerits
        self.previous = previous

    def __repr__(self):
        return '<_BreakNode at %i>' % self.position


class ObjectList(list):

    """Class representing a list of Box, Glue, and Penalty objects.
    Supports the same methods as regular Python lists.
    """
    # Set this to 1 to trace the execution of the algorithm.
    debug = 0

    def __init__(self):
        super().__init__()
        self.sum_width, self.sum_stretch, self.sum_shrink = {}, {}, {}

    def add_closing_penalty(self):
        "Add the standard glue and penalty for the end of a paragraph"
        self.append(Penalty(0, INFINITY, 0))
        self.append(Glue(0, 0, INFINITY))
        self.append(Penalty(0, -INFINITY, 1))

    def is_feasible_breakpoint(self, i):
        "Return true if position 'i' is a feasible breakpoint."
        box = self[i]
        if box.is_penalty and box.penalty < INFINITY:
            return 1
        elif i > 0 and box.is_glue and self[i-1].is_box:
            return 1
        else:
            return 0

    def is_forced_break(self, i):
        "Return true if position 'i' is a forced breakpoint."
        box = self[i]
        if box.is_penalty and box.penalty == -INFINITY:
            return 1
        else:
            return 0

    def measure_width(self, pos1, pos2):
        "Add up the widths between positions 1 and 2"

        return self.sum_width[pos2] - self.sum_width[pos1]

    def measure_stretch(self, pos1, pos2):
        "Add up the stretch between positions 1 and 2"

        return self.sum_stretch[pos2] - self.sum_stretch[pos1]

    def measure_shrink(self, pos1, pos2):
        "Add up the shrink between positions 1 and 2"
        return self.sum_shrink[pos2] - self.sum_shrink[pos1]

    def compute_adjustment_ratio(self, pos1, pos2, line, line_lengths):
        "Compute adjustment ratio for the line between pos1 and pos2"
        length = self.measure_width(pos1, pos2)
        if self[pos2].is_penalty:
            length = length + self[pos2].width
        if self.debug:
            print('\tline length=', length)

        # Get the length of the current line; if the line_lengths list
        # is too short, the last value is always used for subsequent
        # lines.

        if line < len(line_lengths):
            available_length = line_lengths[line]
        else:
            available_length = line_lengths[-1]

        # Compute how much the contents of the line would have to be
        # stretched or shrunk to fit into the available space.
        if length < available_length:
            stretch = self.measure_stretch(pos1, pos2)
            if self.debug:
                print('\tLine too short: shortfall = %i, stretch = %i'
                      % (available_length - length, stretch))
            if stretch > 0:
                ratio = (available_length - length) / float(stretch)
            else:
                ratio = INFINITY

        elif length > available_length:
            shrink = self.measure_shrink(pos1, pos2)
            if self.debug:
                print('\tLine too long: extra = %s, shrink = %s'
                      % (available_length - length, shrink))
            if shrink > 0:
                ratio = (available_length - length) / float(shrink)
            else:
                ratio = INFINITY
        else:
            # Exactly the right length!
            ratio = 0

        return ratio

    def add_active_node(self, active_nodes, node):
        """Add a node to the active node list.
        The node is added so that the list of active nodes is always
        sorted by line number, and so that the set of (position, line,
        fitness_class) tuples has no repeated values.
        """
        index = 0
        # Find the first index at which the active node's line number
        # is equal to or greater than the line for 'node'.  This gives
        # us the insertion point.
        while (index < len(active_nodes) and
               active_nodes[index].line < node.line):
            index = index + 1
        insert_index = index
        # Check if there's a node with the same line number and
        # position and fitness.  This lets us ensure that the list of
        # active nodes always has unique (line, position, fitness)
        # values.
        while (index < len(active_nodes) and
               active_nodes[index].line == node.line):
            if (active_nodes[index].fitness_class == node.fitness_class and
                    active_nodes[index].position == node.position):
                # A match, so just return without adding the node
                return
            index += 1
        active_nodes.insert(insert_index, node)

    def compute_breakpoints(self, line_lengths, looseness=0, tolerance=1,
                            fitness_demerit=100, flagged_demerit=100):
        """Compute a list of optimal breakpoints for the paragraph
        represented by this ObjectList, returning them as a list of
        integers, each one the index of a breakpoint.

        line_lengths : a list of integers giving the lengths of each
                       line.  The last element of the list is reused
                       for subsequent lines.
        looseness : An integer value. If it's positive, the paragraph
                   will be set to take that many lines more than the
                   optimum value.   If it's negative, the paragraph is
                   set as tightly as possible.  Defaults to zero,
                   meaning the optimal length for the paragraph.
        tolerance : the maximum adjustment ratio allowed for a line.
                    Defaults to 1.
        fitness_demerit : additional value added to the demerit score
                          when two consecutive lines are in different
                          fitness classes.
        flagged_demerit : additional value added to the demerit score
                          when breaking at the second of two flagged
                          penalties.
        """
        length = len(self)
        if not len(self):
            return []            # No text, so no breaks

        # Precompute lists containing the numeric values for each box.
        # The variable names follow those in Knuth's description.
        widths, stretches, shrinks = [0] * length, [0] * length, [0] * length
        penalties, flags = [0] * length, [0] * length
        for i in range(length):
            box = self[i]
            widths[i] = box.width
            if box.is_glue:
                stretches[i] = box.stretch
                shrinks[i] = box.shrink
            elif box.is_penalty:
                penalties[i] = box.penalty
                flags[i] = box.flagged

        # Precompute the running sums of width, stretch, and shrink
        # (W,Y,Z in the original paper).  These make it easy to measure the
        # width/stretch/shrink between two indexes; just compute
        # sum_*[pos2] - sum_*[pos1].  Note that sum_*[i] is the total
        # up to but not including the box at position i.
        width_sum = stretch_sum = shrink_sum = 0
        for i in range(length):
            self.sum_width[i] = width_sum
            self.sum_stretch[i] = stretch_sum
            self.sum_shrink[i] = shrink_sum

            box = self[i]
            width_sum += box.width
            stretch_sum += box.stretch
            shrink_sum += box.shrink

        # Initialize list of active nodes to a single break at the
        # beginning of the text.
        a_node = _BreakNode(position=0, line=0, fitness_class=1,
                            totalwidth=0, totalstretch=0,
                            totalshrink=0, demerits=0)
        active_nodes = [a_node]
        for i in range(length):
            breakpoint = self[i]
            # Determine if this box is a feasible breakpoint and
            # perform the main loop if it is.
            if self.is_feasible_breakpoint(i):
                # Loop over the list of active nodes, and compute the fitness
                # of the line formed by breaking at A and B.  The resulting
                breaks = []                 # List of feasible breaks
                for a_node in active_nodes[:]:
                    ratio = self.compute_adjustment_ratio(a_node.position, i,
                                                          a_node.line,
                                                          line_lengths)
                    # XXX is 'or' really correct here?  This seems to
                    # remove all active nodes on encountering a forced break!
                    if ratio < -1 or breakpoint.is_forced_break:
                        # Deactivate node A
                        if len(active_nodes) == 1:
                            if self.debug:
                                print("Can't remove last node!")
                                # XXX how should this be handled?
                                # Raise an exception?
                        else:
                            active_nodes.remove(a_node)

                    if -1 <= ratio <= tolerance:
                        # Compute demerits and fitness class
                        if penalties[i] >= 0:
                            demerits = (1 + 100 * abs(ratio)**3 +
                                        penalties[i]) ** 3
                        elif self.is_forced_break(i):
                            demerits = ((1 + 100 * abs(ratio)**3) ** 2 -
                                        penalties[i]**2)
                        else:
                            demerits = (1 + 100 * abs(ratio)**3) ** 2
                        demerits += (flagged_demerit * flags[i] *
                                     flags[a_node.position])
                        # Figure out the fitness class of this line
                        # (tight, loose, very tight or very loose).
                        if ratio < -.5:
                            fitness_class = 0
                        elif ratio <= .5:
                            fitness_class = 1
                        elif ratio <= 1:
                            fitness_class = 2
                        else:
                            fitness_class = 3
                        # If two consecutive lines are in very
                        # different fitness classes, add to the
                        # demerit score for this break.
                        if abs(fitness_class - a_node.fitness_class) > 1:
                            demerits += fitness_demerit
                        # Record a feasible break from A to B
                        brk = _BreakNode(position=i, line=a_node.line + 1,
                                         fitness_class=fitness_class,
                                         totalwidth=self.sum_width[i],
                                         totalstretch=self.sum_stretch[i],
                                         totalshrink=self.sum_shrink[i],
                                         demerits=demerits,
                                         previous=a_node)
                        breaks.append(brk)
                for brk in breaks:
                    self.add_active_node(active_nodes, brk)
        # Find the active node with the lowest number of demerits.
        a_node = min(active_nodes, key=lambda a_node: a_node.demerits)
        if looseness:
            # The search for the appropriate active node is a bit more
            # complicated; we look for a node with a paragraph length
            # that's as close as possible to (A.line+looseness), and
            # with the minimum number of demerits.
            best = 0
            demerits = INFINITY
            for node in active_nodes:
                delta = node.line - a_node.line
                # The two branches of this 'if' statement
                # are for handling values of looseness that are
                # either positive or negative.
                if looseness <= delta < best or best < delta < looseness:
                    # s = delta
                    demerits = node.demerits
                    breaking_node = node
                elif delta == best and node.demerits < demerits:
                    # This break is of the same length, but has fewer
                    # demerits and hence is a more attractive one.
                    demerits = node.demerits
                    breaking_node = node
            a_node = breaking_node
        # Use the chosen node A to determine the optimum breakpoints,
        # and return the resulting list of breakpoints.
        breaks = []
        while a_node is not None:
            breaks.append(a_node.position)
            a_node = a_node.previous
        breaks.reverse()
        return breaks


# Simple test code.
def main():
    """Test it"""
    text = """Writing this summary was difficult, because there were no large
    themes in the last two weeks of discussion.  Instead there were lots
    and lots of small items; as the release date for 2.0b1 nears, people are
    concentrating on resolving outstanding patches, fixing bugs, and
    making last-minute tweaks.
    Computes a Adler-32 checksum of string.  (An Adler-32
    checksum is almost as reliable as a CRC32 but can be computed much
    more quickly.)  If value is present, it is used as the
    starting value of the checksum; otherwise, a fixed default value is
    used.  This allows computing a running checksum over the
    concatenation of several input strings.  The algorithm is not
    cryptographically strong, and should not be used for
    authentication or digital signatures."""
    text = ' '.join(text.split())

    line_width = 100   # Line width to use for formatting
    full_justify = True   # Boolean; if true, do full justification
    # Turn chunk of text into an ObjectList.
    obj_list = ObjectList()
    for character in text:
        if character in ' \n':
            # Append interword space -- 2 units +/- 1
            obj_list.append(Glue(3, 2, 10))
        elif character == '@':
            # Append forced break
            obj_list.append(Penalty(0, -INFINITY))

        else:
            # All characters are 1 unit wide
            obj_list.append(Box(1, character))

    # Append closing penalty and glue
    obj_list.add_closing_penalty()

    # Compute the breakpoints
    line_lengths = [line_width]
    # line_lengths = range(120, 20, -10)
    breaks = obj_list.compute_breakpoints(line_lengths, tolerance=2)

    assert breaks[0] == 0
    line_start = 0
    line_no = 0
    for breakpoint in breaks[1:]:
        adj_ratio = obj_list.compute_adjustment_ratio(line_start, breakpoint,
                                                      line_no, line_lengths)
        line_no += 1
        for i in range(line_start, breakpoint):
            box = obj_list[i]
            if box.is_glue:
                if full_justify:
                    box_width = int(box.compute_width(adj_ratio))
                else:
                    box_width = 1
                sys.stdout.write(' ' * box_width)

            elif hasattr(box, 'character'):
                sys.stdout.write(box.character)
        line_start = breakpoint + 1
        sys.stdout.write('\n')

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Constants shared between modules in the rpi2caster package"""

# Build a sequence of Monotype signals as they appear on the paper tower
SIGNALS = ([x for x in 'NMLKJIHGFSED'] + ['0075'] + [x for x in 'CBA'] +
           [str(x) for x in range(1, 15)] + ['0005', 'O15'])

# Column numbers in 15x17 diecase
COLUMNS_15 = [x for x in 'ABCDEFGHIJKLMNO']
COLUMNS_17 = ['NI', 'NL'] + COLUMNS_15

# For parsing lines in ribbon files
COMMENT_SYMBOLS = ['**', '*', '//', '##', '#']
ASSIGNMENT_SYMBOLS = ['=', ':', ' ']

# Useful for config files (all aliases must be lowercase)
TRUE_ALIASES = ['true', 'on', 'yes', '1']
FALSE_ALIASES = ['false', 'off', 'no', '0']



# -*- coding: utf-8 -*-
"""Helper routines and classes"""


def singleton(cls):
    """Decorator function for singletons"""
    instances = {}

    def getinstance():
        if cls not in instances:
            instances[cls] = cls()
        return instances[cls]
    return getinstance



# -*- coding: utf-8 -*-
"""
Database:

Database-related classes for rpi2caster suite.
"""
# IMPORTS:
# Used for serializing lists stored in database, and for communicating
# with the web application (in the future):
import json
# this module uses sqlite3 database for storing caster, interface,
# wedge, diecase & matrix parameters:
import sqlite3
# Custom exceptions
from . import exceptions
# Get database paths from configuration
from .global_config import DATABASE_PATHS


class Singleton(type):
    """Make only one object"""
    instance = None

    def __call__(cls, *args, **kw):
        if not cls.instance:
            cls.instance = super(Singleton, cls).__call__(*args, **kw)
        return cls.instance


class Database(object):
    """Database(database_path, confFilePath):

    A class containing all methods related to storing, retrieving
    and deleting data from a SQLite3 database used for config.

    We're using database because it's easy to access and modify with
    third-party programs (like sqlite, sqlitebrowser or a Firefox plugin),
    and there will be lots of data to store: diecase (matrix case)
    properties, diecase layouts, wedge unit values, caster and interface
    settings (although we may move them to config files - they're "system"
    settings best left default, instead of "foundry" settings the user has
    to set up before being able to cast, based on their type foundry's
    inventory, which varies from one place to another).

    Methods here are for reading/writing data for diecases, matrices,
    wedges (and casters & interfaces) from/to designated sqlite3 database.

    Default database path is /var/lib/rpi2caster - but you can
    override it by instantiating this class with a different name
    passed as an argument. It is necessary that the user who's running
    this program for setup has write access to the database file.
    """
    __metaclass__ = Singleton

    def __init__(self):
        # Connect to the database
        for path in DATABASE_PATHS:
            try:
                self.path = path
                self.db_connection = sqlite3.connect(path)
                # End on first successful find
                break
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Continue looping
                pass
        else:
            # Fell off the end of the loop
            raise exceptions.WrongConfiguration('Cannot connect to '
                                                'database file %s' % path)

    def __repr__(self):
        return self.path

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def get_all_diecases(self):
        """Gets all diecases stored in database."""
        with self.db_connection:
            try:
                sql_query = 'SELECT * FROM matrix_cases ORDER BY diecase_id'
                cursor = self.db_connection.cursor()
                cursor.execute(sql_query)
                results = cursor.fetchall()
                # Check if we got any:
                if not results:
                    raise exceptions.NoMatchingData
                return results
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def add_diecase(self, diecase):
        """Registers a diecase in our database."""
        # data - a list with diecase parameters to be written,
        # layout is a JSON-dumped dictionary
        data = [diecase.diecase_id, diecase.typeface, diecase.wedge.name,
                json.dumps(diecase.layout)]
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                # Create the table first:
                cursor.execute('CREATE TABLE IF NOT EXISTS matrix_cases ('
                               'diecase_id TEXT UNIQUE PRIMARY KEY, '
                               'typeface TEXT NOT NULL, '
                               'wedge TEXT NOT NULL, '
                               'layout TEXT NOT NULL)')
                # Then add an entry:
                cursor.execute('INSERT OR REPLACE INTO matrix_cases ('
                               'diecase_id, typeface, wedge, layout'
                               ') VALUES (?, ?, ?, ?)''', data)
                self.db_connection.commit()
                return True
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def get_diecase(self, diecase_id):
        """Searches for diecase metadata, based on the unique diecase ID."""
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                sql = 'SELECT * FROM matrix_cases WHERE diecase_id = ?'
                cursor.execute(sql, [diecase_id])
                row = cursor.fetchone()
                if not row:
                    raise exceptions.NoMatchingData
                diecase = list(row)
                # De-serialize the diecase layout, convert it back to a list
                raw_layout = json.loads(diecase.pop())
                # Make sure the layout has specified units
                layout = []
                for record in raw_layout:
                    (char, styles, coordinates, units) = record
                    try:
                        units = int(units)
                    except (IndexError, ValueError):
                        units = 0
                    record = (char, styles, coordinates, units)
                    layout.append(record)
                diecase.append(layout)
                return diecase
            except (TypeError, ValueError, IndexError):
                # No data or cannot process it
                raise exceptions.NoMatchingData
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def delete_diecase(self, diecase):
        """Deletes a diecase with given unique ID from the database."""
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                cursor.execute('DELETE FROM matrix_cases WHERE diecase_id = ?',
                               [diecase.diecase_id])
                return True
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def get_all_ribbons(self):
        """Gets all ribbons stored in database"""
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                cursor.execute('SELECT * FROM ribbons ORDER BY ribbon_id')
                # Check if we got any:
                results = cursor.fetchall()
                if not results:
                    raise exceptions.NoMatchingData
                return results
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def add_ribbon(self, ribbon):
        """Registers a ribbon in our database."""
        data = [ribbon.description, ribbon.customer, ribbon.diecase_id,
                ribbon.wedge_name, json.dumps(ribbon.contents)]
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                # Create the table first:
                cursor.execute('CREATE TABLE IF NOT EXISTS ribbons ('
                               'ribbon_id INTEGER PRIMARY KEY AUTOINCREMENT, '
                               'description TEXT, '
                               'customer TEXT, '
                               'diecase_id TEXT NOT NULL, '
                               'wedge TEXT, '
                               'contents TEXT NOT NULL)')
                # Then add an entry:
                cursor.execute('INSERT OR REPLACE INTO ribbons ('
                               'title, author, customer, diecase_id, contents'
                               ') VALUES (?, ?, ?, ?, ?)''', data)
                self.db_connection.commit()
                return True
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def get_ribbon(self, ribbon_id):
        """Searches for ribbon, based on the unique ribbon ID."""
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                cursor.execute('SELECT * FROM ribbons WHERE ribbon_id = ?',
                               [ribbon_id])
                ribbon = list(cursor.fetchone())
                # Take the last item which is contents
                # De-serialize it, convert it back to a list
                raw_contents = json.loads(ribbon.pop())
                # Add the contents
                ribbon.append(raw_contents)
                # Return the list
                return ribbon
            except (TypeError, ValueError, IndexError):
                # No data or cannot process it
                raise exceptions.NoMatchingData
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def delete_ribbon(self, ribbon):
        """Deletes a ribbon entry from database."""
        # Okay, proceed:
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                cursor.execute('DELETE FROM ribbons WHERE ribbon_id = ?',
                               [ribbon.ribbon_id])
                return True
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError



# -*- coding: utf-8 -*-
"""database_backends: connections for various database types,
currently supported: sqlite and postgres"""
import getpass


class GenericDBConnection(object):
    """Generic methods and attributes for connection objects"""
    def __init__(self):
        self.db_conn = None

    def __enter__(self):
        return self.db_conn


class SQLiteConnection(GenericDBConnection):
    """SQLite3 connection - local database on a file"""
    def __init__(self, db_path):
        super().__init__()
        import sqlite3
        try:
            self.db_conn = sqlite3.connect(db_path)
        except (sqlite3.OperationalError, sqlite3.DatabaseError):
            pass


class PostgresConnection(GenericDBConnection):
    """PostgreSQL connection without SSH tunnelling or SSL security"""
    def __init__(self, db_name, db_host='localhost', db_port=5432,
                 db_username='postgres', db_password='postgres'):
        super().__init__()
        import psycopg2
        try:
            self.db_conn = psycopg2.connect(database=db_name,
                                            port=db_port,
                                            user=db_username,
                                            password=db_password,
                                            host=db_host)
        except (psycopg2.OperationalError, psycopg2.DatabaseError):
            pass


class SSLTunnelPostgresConnection(GenericDBConnection):
    """Postgres connection tunnelled over SSH"""
    def __init__(self, db_name, host_address, ssh_port=22,
                 ssh_username=getpass.getuser(), ssh_password='',
                 db_username='postgres',
                 db_password='postgres'):
        super().__init__()
        import psycopg2
        from sshtunnel import SSHTunnelForwarder
        # Connect over SSH tunnel
        listen_addr = ('127.0.0.1', 5432)
        with SSHTunnelForwarder(host_address, ssh_port,
                                ssh_password=ssh_password,
                                ssh_username=ssh_username,
                                remote_bind_address=listen_addr) as server:
            try:
                self.db_conn = psycopg2.connect(database=db_name,
                                                port=server.local_bind_port,
                                                user=db_username,
                                                password=db_password)
            except (psycopg2.OperationalError, psycopg2.DatabaseError):
                pass
